"""
Define configuration defaults for Django project settings.
"""

from collections.abc import Mapping, Sequence

default_health_check_apps: Sequence[str] = [
    "health_check",
    "health_check.db",
    "health_check.cache",
    "health_check.contrib.migrations",
]
"""
The default health check app and plugins to enable.

This set of plugins is configured because they're 99% guaranteed to be used in every
project. Other contrib plugins are omitted because they require more configuration for
which we cannot easily provide defaults.

See https://codingjoe.dev/django-health-check/install/ for more details.
"""

default_health_check_subsets: Mapping[str, Sequence[str]] = {
    # deliberately empty - super cheap check to run that verifies HTTP traffic is
    # working
    "livez": (),
    # light-weight checks - if any of these fail, the instance should not accept live
    # traffic
    "readyz": (
        "DatabaseBackend",
        "MigrationsHealthCheck",
        "Cache backend: default",  # 'default' is the cache alias in ``settings.CACHES``
    ),
}
"""
The default subsets of checks to run. Each subset is named (using the keys of the dict).

The liveness check is intended for the liveness probe in Kubernetes. It can also be used
for the readiness probe, but the readiness check does a slightly more thorough check
without becoming expensive.

The startup probes can use the main health check without specifying any subset, which
checks all the plugins, including possibly expensive checks.

See https://codingjoe.dev/django-health-check/container/#subsets for more details
"""



from django.apps import AppConfig

from .probes import connect_beat_signals, connect_worker_signals


class CeleryHealthChecksAppConfig(AppConfig):
    name = "maykin_common.health_checks.celery"

    def ready(self):
        connect_worker_signals()
        connect_beat_signals()



"""
Command line companion to the ``maykin-common`` Django utilities.

The command line script is deliberately written in pure Python without loading Django
at all (for performance), as opposed to providing Django management commands. Large
Django projects may see slow startup times in the order of 2-10s due to (expensive)
imports when loading all the code. This pairs very badly with health checks machinery
which tend to have timeouts of a couple of seconds.
"""

import importlib.metadata
import socket
import time
from pathlib import Path
from typing import Annotated
from urllib.parse import urlparse, urlunparse

import requests
import typer

app = typer.Typer()

_WORKER_EXIT_CODE_EVENT_LOOP_BROKEN = 1
_WORKER_EXIT_CODE_PING_FAILURE = 2
_WORKER_EXIT_CODE_NOT_READY = 4


@app.command()
def version():
    version = importlib.metadata.version("maykin_common")
    typer.echo(f"maykin-common v{version}")


@app.command(name="health-check")
def health_check(
    endpoint: Annotated[
        str,
        typer.Option(help="Endpoint/path to test for connection and status code."),
    ] = "/_healthz/livez/",
    timeout: Annotated[
        int,
        typer.Option(help="Timeout for the GET request (in seconds)."),
    ] = 3,
):
    """
    Execute an HTTP health check call against the provided endpoint.

    If no host or domain is provided with the ``endpoint`` option, a default of
    ``http://localhost:8000`` will be used.
    """

    # URLs must start with a scheme, otherwise urlparse chokes :-)
    if not (endpoint.startswith("http://") or endpoint.startswith("https://")):
        endpoint = f"http://{endpoint}"

    parsed = urlparse(endpoint)
    normalized_url = urlunparse(
        (
            parsed.scheme,
            parsed.netloc or "localhost:8000",
            parsed.path or "/_healthz/livez/",
            parsed.params,
            parsed.query,
            parsed.fragment,
        )
    )

    try:
        response = requests.get(normalized_url, timeout=timeout)
    except requests.RequestException as exc:
        typer.secho(f"DOWN ({exc.__class__.__name__})", fg=typer.colors.RED, err=True)
        exit(1)

    if up := response.ok:
        typer.secho(
            f"UP, response status code: {response.status_code}",
            fg=typer.colors.GREEN,
        )
    else:
        typer.secho(
            f"DOWN, response status code: {response.status_code}",
            fg=typer.colors.RED,
            err=True,
        )

    exit_code = 0 if up else 1
    exit(exit_code)


@app.command(name="worker-health-check")
def worker_health_check(
    # liveness
    liveness_file: Annotated[
        Path,
        typer.Option(
            help="The event loop liveness file, created and updated by the "
            "`EventLoopProbe`."
        ),
    ] = Path("/tmp") / "celery_worker_event_loop_live",
    max_age: Annotated[
        int,
        typer.Option(
            help="How long ago the last update of the event loop liveness file is "
            "allowed to be, in seconds. You should match this to the "
            "'MKN_HEALTH_CHECKS_WORKER_EVENT_LOOP_PROBE_FREQUENCY_SECONDS' setting."
        ),
    ] = 60 + 10,
    skip_event_loop_liveness: Annotated[
        bool,
        typer.Option(
            help="Opt-out from the event loop liveness check.",
        ),
    ] = False,
    broker: Annotated[
        str,
        typer.Option(
            help="Broker URL, should match the 'CELERY_BROKER' setting. Used for the "
            "ping roundtrip check."
        ),
    ] = "redis://localhost:6379/0",
    worker_name: Annotated[
        str,
        typer.Option(
            envvar="CELERY_WORKER_NAME",
            help="Worker name, typically composed from <queue>@<host>.",
        ),
    ] = f"celery@{socket.gethostname()}",
    ping_timeout: Annotated[
        int, typer.Option(help="Timeout in seconds after which the ping check fails.")
    ] = 3,
    skip_ping: Annotated[
        bool, typer.Option(help="Opt-out from the ping roundtrip check.")
    ] = False,
    # readiness
    readiness_file: Annotated[
        Path,
        typer.Option(
            help="The readiness file, created when the worker is ready to process "
            "tasks."
        ),
    ] = Path("/tmp") / "celery_worker_ready",
    skip_readiness: Annotated[
        bool, typer.Option(help="Opt-in to the readiness check.")
    ] = True,
):
    """
    Run health checks for the Celery worker.

    The worker health checks consist of separate components. The defaults are geared
    towards liveness checks, but you can disable/enable parts to adapt to your
    situation.

    * Check the `liveness-file` to detect issues in the worker event loop. If the event
      loop is broken, the worker will definitely not be processing tasks and a restart
      is necessary.
    * Ping the worker using Celery's inspection machinery. Pinging verifies the broker
      connection, as it performs a complete roundtrip. Broker connection loss can
      sometimes be solved by restarting the worker.
    * Check the presence of the `readiness-file`. Absence of the readiness file
      indicates the worker is not (yet) ready to process tasks. It may have been stopped
      or is still starting up.

    If any check fails, the command exist with a non-zero exit code.
    """
    # always instantiate an app as a sanity check
    try:
        from celery import Celery
    except ImportError:  # pragma: no cover
        typer.secho(
            "Could not import celery - please make sure to execute this in a celery"
            "worker environment.",
            fg=typer.colors.RED,
            err=True,
        )
        exit(1)

    celery_app = Celery(broker=broker)

    if not skip_event_loop_liveness:
        if not liveness_file.exists() or not liveness_file.is_file():
            typer.secho(
                f"File '{liveness_file}' does not exist or is not a file.",
                fg=typer.colors.RED,
                err=True,
            )
            exit(_WORKER_EXIT_CODE_EVENT_LOOP_BROKEN)

        now = time.time()
        last_modified = liveness_file.stat().st_mtime
        age_in_seconds = int(now - last_modified)
        if age_in_seconds > max_age:
            typer.secho(
                f"File '{liveness_file}' is older than max-age.",
                fg=typer.colors.RED,
                err=True,
            )
            exit(_WORKER_EXIT_CODE_EVENT_LOOP_BROKEN)
        else:
            typer.secho(
                "The event loop appears to be running.",
                fg=typer.colors.GREEN,
            )

    if not skip_ping:
        replies = celery_app.control.ping(
            destination=[worker_name], timeout=ping_timeout
        )
        if replies:
            typer.secho(f"{worker_name}: PONG.", fg=typer.colors.GREEN)
        else:
            typer.secho(
                f"No reply to ping from '{worker_name}' after {ping_timeout}s.",
                fg=typer.colors.RED,
                err=True,
            )
            exit(_WORKER_EXIT_CODE_PING_FAILURE)

    if not skip_readiness:
        if not readiness_file.exists() or not readiness_file.is_file():
            typer.secho(
                f"File '{readiness_file}' does not exist - worker is not ready.",
                fg=typer.colors.RED,
                err=True,
            )
            exit(_WORKER_EXIT_CODE_NOT_READY)
        else:
            typer.secho(
                "The worker appears ready to process tasks.",
                fg=typer.colors.GREEN,
            )

    exit(0)


@app.command(name="beat-health-check")
def beat_health_check(
    file: Annotated[
        Path,
        typer.Option(
            help="The liveness file, created and updated by the Beat health check "
            "machinery."
        ),
    ] = Path("/tmp") / "celery_beat_live",
    max_age: Annotated[
        int,
        typer.Option(
            help="How long ago the last update of liveness file is allowed to be, in "
            "seconds. You should tune this to the beat schedule of your application."
        ),
    ] = 3600,
):
    """
    Check the last modified timestamp of the Beat liveness file.

    If it's older than ``max-age``, Beat is considered unhealthy.
    """
    file = file.resolve()
    if not file.exists() or not file.is_file():
        typer.secho(
            f"File '{file}' does not exist or is not a file.",
            fg=typer.colors.RED,
            err=True,
        )
        exit(1)

    # check the file age
    now = time.time()
    last_modified = file.stat().st_mtime
    age_in_seconds = int(now - last_modified)
    if age_in_seconds > max_age:
        typer.secho(
            f"File '{file}' is older than max-age.",
            fg=typer.colors.RED,
            err=True,
        )
        exit(1)
    else:
        human_readable_age: str = f"{age_in_seconds}s"
        if 60 < age_in_seconds < 3600:
            age_in_minutes = round(age_in_seconds / 60, 1)
            human_readable_age = f"{age_in_minutes:.1f}m".replace(".0", "")
        elif age_in_seconds >= 3600:
            age_in_hours = round(age_in_seconds / 3600, 1)
            human_readable_age = f"{age_in_hours:.1f}h".replace(".0", "")

        typer.secho(
            f"Last scheduled task: {human_readable_age} ago.",
            fg=typer.colors.GREEN,
        )
        exit(0)


if __name__ == "__main__":  # pragma: no cover
    app()



import atexit
import logging
from pathlib import Path
from typing import TYPE_CHECKING

from celery import bootsteps
from celery.apps.worker import Worker as _Worker
from celery.beat import Service as BeatService
from celery.signals import after_task_publish, beat_init, worker_ready, worker_shutdown
from celery.worker.consumer import Consumer
from kombu.asynchronous.timer import Entry as TimerEntry, Timer

from maykin_common.settings import get_setting

logger = logging.getLogger(__name__)

if TYPE_CHECKING:

    class Worker(_Worker):
        """
        Worker class with timer attribute.

        Assigned through :meth:`celery.worker.components.Timer.create`, it's a dynamic
        attribute that's not present in the base class definitions, so we have to
        massage it a bit.
        """

        timer: Timer
else:
    Worker = _Worker


#
# Utilities for checking the health of celery worker
#


class EventLoopProbe(bootsteps.StartStopStep):
    """
    Checks that the Celery worker event loop is alive.

    When the Celery worker starts, it starts the event loop, timer and processing pool.
    As a "final" step, it starts the Consumer blueprint, which is responsible for
    establishing the broker connection and actually start processing/consuming messages
    and tasks.

    This event loop probe installs a bootstep when the timer is available, and schedules
    a periodic callback that touches a liveness file. If the timestamp of the last
    modified moment of the liveness file is too long again, we can conclude/assume that
    the event loop has crashed and the worker should be restarted, as it's likely that
    ETA/countdown tasks and tasks in general are not being processed anymore by this
    worker. This makes no guarantees about actually being able to consume tasks or a
    live connection though. The timer runs in the main worker process (when using
    the preforking/multi-processing pool).

    Celery itself *should* re-establish broker connection on connection loss, by
    restarting the Consumer blueprint, but bugs in Celery itself have been observed in
    the past. We can implement connectivity checks by pinging the worker from itself,
    which is set up elsewhere.

    See the `upstream <https://docs.celeryq.dev/en/stable/userguide/extending.html#blueprints>`_
    documentation for details about blueprints and bootstep mechanisms.

    Usage::

        >>> app = Celery("my-project")
        >>> app.steps["worker"].add(EventLoopProbe)
    """

    # we need the Timer component before we can run this bootstep. Celery uses this to
    # figure out the step dependency graph.
    requires = {"celery.worker.components:Timer"}
    tref: TimerEntry | None = None
    liveness_file: Path

    def start(self, parent: Worker):
        self.liveness_file = liveness_file = get_setting(
            "MKN_HEALTH_CHECKS_WORKER_EVENT_LOOP_LIVENESS_FILE"
        )
        # create intermediate directories if they don't yet exist
        if not (parent_dir := liveness_file.parent).exists():
            parent_dir.mkdir(parents=True, exist_ok=True)

        liveness_file.touch()
        frequency: int = get_setting(
            "MKN_HEALTH_CHECKS_WORKER_EVENT_LOOP_PROBE_FREQUENCY_SECONDS"
        )
        self.tref = parent.timer.call_repeatedly(
            frequency, liveness_file.touch, priority=10
        )

    def stop(self, parent: Worker):
        assert self.tref is not None
        self.tref.cancel()
        self.liveness_file.unlink(missing_ok=True)


def on_worker_ready(*, sender: Consumer, **kwargs):
    """
    Create/touch the readiness file when the worker is ready to accept work.
    """
    readiness_file: Path = get_setting("MKN_HEALTH_CHECKS_WORKER_READINESS_FILE")
    # create intermediate directories if they don't yet exist
    if not (parent_dir := readiness_file.parent).exists():
        parent_dir.mkdir(parents=True, exist_ok=True)
    readiness_file.touch()
    logger.info("worker_ready")


def on_worker_shutdown(*, sender: Worker, **kwargs):
    """
    Delete the readiness file when a worker shuts down.
    """
    logger.info("worker_shutdown")
    readiness_file: Path = get_setting("MKN_HEALTH_CHECKS_WORKER_READINESS_FILE")
    readiness_file.unlink(missing_ok=True)


def connect_worker_signals():
    worker_ready.connect(on_worker_ready, dispatch_uid="probes.on_worker_ready")
    worker_shutdown.connect(
        on_worker_shutdown, dispatch_uid="probes.on_worker_shutdown"
    )


#
# Utilities for checking the health of celery beat
#

_RUNNING_IN_BEAT = False


def on_beat_init(*, sender: BeatService, **kwargs):
    global _RUNNING_IN_BEAT
    _RUNNING_IN_BEAT = True
    logger.debug("beat_process_marked")
    liveness_file: Path = get_setting("MKN_HEALTH_CHECKS_BEAT_LIVENESS_FILE")
    # on shutdown, clear up the liveness file
    atexit.register(liveness_file.unlink, missing_ok=True)


def on_beat_task_published(*, sender: str, routing_key: str, **kwargs):
    """
    Update the celery beat liveness every time a task is successfully published.

    ``after_task_publish`` fires in the process that sent the task, so we must discern
    between the regular Django app that schedules tasks, and celery beat that also
    schedules tasks. We do this by tapping into the ``beat_init`` signal to mark the
    process as a beat process, and only touch the liveness file when running in beat.
    """
    if not _RUNNING_IN_BEAT:
        return

    liveness_file: Path = get_setting("MKN_HEALTH_CHECKS_BEAT_LIVENESS_FILE")
    logger.debug(
        "beat_task_published", extra={"task": sender, "routing_key": routing_key}
    )
    # create intermediate directories if they don't yet exist
    if not (parent_dir := liveness_file.parent).exists():
        parent_dir.mkdir(parents=True, exist_ok=True)
    # touching the file updates the last modified timestamp, which can be checked by
    # the health-check command
    liveness_file.touch()


def connect_beat_signals():
    # register signals for beat so that we can health-check it
    beat_init.connect(on_beat_init, dispatch_uid="probes.on_beat_init")
    after_task_publish.connect(
        on_beat_task_published, dispatch_uid="probes.on_beat_task_published"
    )



"""
Utilities to read and process configuration for a project.

.. todo:: Incorporate the Team Bron documentation options for parameters.
"""

from __future__ import annotations

import csv
import io
from collections.abc import Callable, Sequence
from typing import Literal, Never, assert_never, overload

from decouple import Csv, Undefined, config as _config, undefined

__all__ = ["config"]


@overload
def config(option: str) -> str: ...


@overload
def config[T](
    option: str,
    *,
    default: Sequence[T],
    split: Literal[True],
    cast: Callable[[str], T] | Undefined = undefined,
) -> list[T]: ...


@overload
def config(
    option: str,
    *,
    default: Undefined = undefined,
    split: Literal[True],
    cast: Undefined = undefined,
) -> list[str]: ...


@overload
def config[T](
    option: str,
    *,
    default: Undefined = undefined,
    split: Literal[True],
    cast: Callable[[str], T],
) -> list[T]: ...


@overload
def config(option: str, *, default: None) -> str | None: ...


@overload
def config[T](option: str, *, default: T | Undefined = undefined) -> T: ...


# because we can't express difference / negation types: object \ None
@overload
def config(option: str, *, default: None, cast: Callable) -> Never: ...


@overload
def config[T](
    option: str, *, default: str | Undefined = undefined, cast: Callable[[str], T]
) -> T: ...


def config[T](
    option: str,
    *,
    default: T | Sequence[T] | None | str | Undefined = undefined,
    split: bool = False,
    cast: Callable[[str], T] | Undefined = undefined,
) -> str | None | T | Sequence[T]:
    """
    Pull a config parameter from the environment.

    Read the config variable ``option``. If it's optional, use the ``default`` value.

    If not ``cast`` parameter is provided, then the ``cast`` is derived from the
    ``default`` type when a default is provided. However, when you provide a ``cast``
    parameter explicitly, you must provide any ``default`` as a string as it will be
    passed to the provided ``cast`` callback.

    Note that ``default=None`` does not mean there's no default;
    omitting the ``default`` kwarg entirely means there's no default.

    Pass ``split=True`` to split the comma-separated input into a list. If a default is
    provided, it must be a list.

    Examples::

        >>> SECRET_KEY: str = config("SECRET_KEY")
        >>> DB_NAME: str = config("DB_NAME", default="my-awesome-project")
        >>> DB_PORT: int = config("DB_PORT", default=5432")
        >>> SESSION_COOKIE_DOMAIN: str | None = config(
        ...     "SESSION_COOKIE_DOMAIN", default=None
        ... )
        >>> ALLOWED_HOSTS: list[str] = config("ALLOWED_HOSTS", split=True, default=[])
        >>> CUSTOM = config(
        ...     "CUSTOM",
        ...     default="123",
        ...     cast=lambda v: int(v) if v is not None else None,
        ... )  # typed as int | None
    """

    if split:
        assert isinstance(default, Undefined | Sequence), (
            "You must provide a sequence default argument"
        )
        match default:
            case [t, *_]:
                return _config(
                    option,
                    cast=Csv(cast=cast if callable(cast) else type(t)),
                    default=_dumps(default),
                )
            case []:
                return _config(
                    option,
                    cast=Csv(cast=cast if callable(cast) else lambda x: x),
                    default="",
                )
            case _:
                if callable(cast):
                    return _config(option, cast=Csv(cast=cast))
                return _config(option, cast=Csv())

    # infer the ``cast`` from the default if not provided explicitly
    match (cast, default):
        #
        # cases without explicit cast
        #
        case Undefined(), Undefined():
            return _config(option)
        case Undefined(), None:
            # explicit `None` default values cannot be used as cast, ignore it.
            return _config(option, default=default)
        case Undefined(), _:
            return _config(option, default=default, cast=type(default))
        #
        # with explicit cast
        #
        case _, Undefined():
            return _config(option, cast=cast)
        case _, _:
            # the combination of a default + cast is odd and goes against the common
            # behaviour when it's derived from the default type - unfortunately we can't
            # simply take the ``str(default)``, as there's no guarantee that
            # ``default == cast(str(default))`` (e.g. ``cast=date.fromisoformat``
            # already falls apart)
            if not isinstance(default, str):
                raise TypeError(
                    "'default' must be a string when providing a cast callback"
                )
            return _config(option, default=default, cast=cast)
        case _:  # pragma: no cover
            assert_never((cast, default))


def _dumps(given_default: Sequence) -> str:
    # python-decouple Csv cast expects the default as a string -
    # serialize it again.
    fd = io.StringIO()
    csv.writer(fd, lineterminator="", quoting=csv.QUOTE_NONNUMERIC).writerow(
        given_default
    )
    fd.seek(0)
    default = fd.read()
    return default



from django.urls import include, path

# Default/convention for URL patterns. With all the defaults, this makes the following
# URLs available:
#
# * ``/_healthz/`` -> reports on all health checks configured
# * ``/_healthz/livez/`` -> no plugins at all, simple check if the app is alive
# * ``/_healthz/readyz/`` -> essential plugins, check if the app can do useful work
urlpatterns = [
    path("_healthz/", include("health_check.urls")),
]



"""
Defaults and conventions around `django-health-check`_.

Out of the box, we support health checks that verify:

* request-response cycles function
* the application can read and write from/to the database
* there are no pending database migrations
* the (default) cache is functioning

On an individual project basis, you can extend the application with additional checks
and implement your own - see the `upstream documentation`_.

.. _django-health-check: https://pypi.org/project/django-health-check/
.. _upstream documentation: https://codingjoe.dev/django-health-check/
"""

import importlib.util

from django.core.exceptions import ImproperlyConfigured

if importlib.util.find_spec("health_check") is None:  # pragma: no cover
    raise ImproperlyConfigured(
        "You must install the health-checks extra to use health checks."
    )

from .defaults import default_health_check_apps, default_health_check_subsets

__all__ = [
    "default_health_check_apps",
    "default_health_check_subsets",
]



from __future__ import annotations

from opentelemetry.context import Context
from opentelemetry.sdk.trace import Span, SpanProcessor


class CustomAttributeSpanProcessor(SpanProcessor):
    def on_start(self, span: Span, parent_context: Context | None = None) -> None:
        """
        Determine type and subtype based on the active instrumentation library.
        """
        if span.attributes and span.attributes.get("span.type"):
            return

        instrumentation_scope = span.instrumentation_scope
        library_name = (
            instrumentation_scope.name.lower() if instrumentation_scope else ""
        )

        span_type: str
        span_subtype: str

        match library_name:
            case "opentelemetry.instrumentation.redis":
                span_type = "db"
                span_subtype = "redis"
            case "opentelemetry.instrumentation.psycopg":
                span_type = "db"
                span_subtype = "postgresql"
            case "opentelemetry.instrumentation.django":
                span_type = (
                    "web"
                    if span.attributes and span.attributes.get("http.method")
                    else "app"
                )
                span_subtype = "django"
            case "opentelemetry.instrumentation.requests":
                span_type = "external"
                span_subtype = "http"
            case "opentelemetry.instrumentation.celery":
                span_type = "async"
                span_subtype = "celery"
            case _:
                span_type = "unknown"
                span_subtype = "unknown"

        span.set_attribute("span.type", span_type)
        span.set_attribute("span.subtype", span_subtype)



import itertools
from collections.abc import Collection

from opentelemetry.metrics import Observation


class MetricsAssertMixin:
    def assertMarkedGlobal(self, result: Collection[Observation]) -> None:
        self.assertTrue(  # pyright: ignore[reportAttributeAccessIssue]
            all(
                observation.attributes and observation.attributes["scope"] == "global"
                for observation in result
            ),
            "Global metrics must be labelled as such",
        )

    @staticmethod
    def _group_observations_by(
        observations: Collection[Observation], attribute: str
    ) -> dict[str, int | float]:
        def _key_func(o: Observation) -> str:
            assert o.attributes is not None
            form_name = o.attributes[attribute]
            assert isinstance(form_name, str)
            return form_name

        grouped_by_attribute = itertools.groupby(
            sorted(observations, key=_key_func), key=_key_func
        )
        return {
            form_name: sum(o.value for o in _observations)
            for form_name, _observations in grouped_by_attribute
        }



import importlib.util

from django.core.exceptions import ImproperlyConfigured

if importlib.util.find_spec("celery") is None:  # pragma: no cover
    raise ImproperlyConfigured("You must install celery to use celery health checks.")



from django.views.generic import TemplateView


class ComponentIndexView(TemplateView):
    template_name = "index_component.html"

    component = ""
    api_version = ""
    notification_url = ""

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context.update(
            {
                "component": self.component,
                "api_version": self.api_version,
                "notification_url": self.notification_url,
            }
        )
        return context



from django.contrib.auth import views as auth_views

from maykin_common.throttling import IPThrottleMixin


class PasswordResetView(IPThrottleMixin, auth_views.PasswordResetView):
    throttle_name = "password-reset"
    throttle_visits = 5
    throttle_period = 60
    throttle_methods = ("get",)



from pathlib import Path
from typing import Literal

from django.conf import settings

GOOGLE_ANALYTICS_ID = None
"""
Google analytics id added to the request context.
"""

SHOW_ENVIRONMENT = False
"""
Enable or disable the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_BACKGROUND_COLOR = "orange"
"""
The background color of the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_FOREGROUND_COLOR = "black"
"""
The foreground color of the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT_LABEL = None
"""
The textual content of the ``{% environment_info %}`` template tag.
"""

ENVIRONMENT = ""
"""
The deployment environment label, e.g. 'staging' or 'prod'.

This setting is included in the OpenTelemetry resource attributes.
"""

RELEASE = None
"""
The release version shown in the ``{% version_info %}`` template tag.

This setting is included in the OpenTelemetry resource attributes.
"""

GIT_SHA = None
"""
The commit hash shown in the ``{% version_info %}`` template tag.
"""

PDF_BASE_URL_FUNCTION = None
"""
Function that returns the base url needed to download/resolve custom fonts and/or any
image URLs included in the document to render.

Required for the :ref:`quickstart_pdf` extra.
"""

LOGIN_URLS = []
"""
Collection of login URLs.

This setting is used in the :func:`maykin_common.views.csrf_failure` view to handle CSRF
errors that occur when attempting to log in a second time.
"""

MKN_HEALTH_CHECKS_BEAT_LIVENESS_FILE: Path = Path("/tmp") / "celery_beat_live"
"""
Path to the file that acts as a liveness marker of Celery Beat.

Used when :mod:`maykin_common.health_checks.celery` is installed, which registers health
check mechanisms for Celery Beat. The intermediate directories will be created if
absent. The file itself will be created when beat publishes its first task, and will
be unlinked again when beat exits.
"""

MKN_HEALTH_CHECKS_WORKER_EVENT_LOOP_LIVENESS_FILE: Path = (
    Path("/tmp") / "celery_worker_event_loop_live"
)
"""
Path to the file that acts as a liveness marker of the Celery Worker event loop.

Used when the :class:`maykin_common.health_checks.celery.probes.EventLoopProbe` is added
as a Celery Worker bootstep, which periodically updates the liveness file with the last
timer tick. The intermediate directories will be created if absent. The file itself will
be created when the worker main process starts, and will be unlinked again when the
worker exits.
"""

MKN_HEALTH_CHECKS_WORKER_EVENT_LOOP_PROBE_FREQUENCY_SECONDS: int = 60
"""
Frequency in seconds on how often the event loop should update the liveness file.
"""

MKN_HEALTH_CHECKS_WORKER_READINESS_FILE: Path = Path("/tmp") / "celery_worker_ready"
"""
Path to the file that acts as a readiness marker of Celery Worker.

Used when :mod:`maykin_common.health_checks.celery` is installed, which registers health
check mechanisms for Celery Worker. The intermediate directories will be created if
absent. The file itself will be created when the worker is ready to accept work, i.e.
once the consumer blueprint is finished as part of the celery worker startup. It means
that a connection with the broker is established and tasks can be processed.

When the worker shuts down, the file is unlinked.
"""

type SettingName = Literal[
    "GOOGLE_ANALYTICS_ID",
    "ENVIRONMENT",
    "SHOW_ENVIRONMENT",
    "ENVIRONMENT_BACKGROUND_COLOR",
    "ENVIRONMENT_FOREGROUND_COLOR",
    "ENVIRONMENT_LABEL",
    "RELEASE",
    "GIT_SHA",
    "PDF_BASE_URL_FUNCTION",
    "LOGIN_URLS",
    "MKN_HEALTH_CHECKS_BEAT_LIVENESS_FILE",
    "MKN_HEALTH_CHECKS_WORKER_EVENT_LOOP_LIVENESS_FILE",
    "MKN_HEALTH_CHECKS_WORKER_EVENT_LOOP_PROBE_FREQUENCY_SECONDS",
    "MKN_HEALTH_CHECKS_WORKER_READINESS_FILE",
]


def get_setting(name: SettingName):
    default = globals()[name]
    return getattr(settings, name, default)



from django import http
from django.conf import settings
from django.http import HttpRequest, HttpResponse, HttpResponseRedirect
from django.template import TemplateDoesNotExist, loader
from django.views.csrf import (
    CSRF_FAILURE_TEMPLATE_NAME,
    csrf_failure as original_csrf_failure,
)
from django.views.decorators.csrf import requires_csrf_token
from django.views.defaults import ERROR_500_TEMPLATE_NAME

from maykin_common.settings import get_setting


@requires_csrf_token
def server_error(
    request: HttpRequest, template_name: str = ERROR_500_TEMPLATE_NAME
) -> http.HttpResponseServerError:
    """
    500 error handler.
    """

    try:
        template = loader.get_template(template_name)
    except TemplateDoesNotExist:
        if template_name != ERROR_500_TEMPLATE_NAME:
            # Reraise if it's a missing custom template.
            raise
        return http.HttpResponseServerError(
            b"<h1>Server Error (500)</h1>", content_type="text/html"
        )
    context = {"request": request}
    return http.HttpResponseServerError(template.render(context))


def csrf_failure(
    request: HttpRequest,
    reason: str = "",
    template_name: str = CSRF_FAILURE_TEMPLATE_NAME,
) -> HttpResponse:
    """
    Catch CSRF failure when trying to login a second time, when already logged
    in, by redirecting to the LOGIN_REDIRECT_URL.
    """
    if request.path in get_setting("LOGIN_URLS") and request.user.is_authenticated:
        return HttpResponseRedirect(settings.LOGIN_REDIRECT_URL)
    return original_csrf_failure(request, reason=reason, template_name=template_name)



from django.core.cache import DEFAULT_CACHE_ALIAS, caches
from django.core.management import BaseCommand


class Command(BaseCommand):
    help = "Clears the Django cache"

    def add_arguments(self, parser):
        parser.add_argument("--alias", help="Clear given cache only")

    def handle(self, *args, **options):
        alias = options["alias"]
        cache = caches[alias] if alias else caches[DEFAULT_CACHE_ALIAS]

        cache.clear()
